@@ -28,10 +28,26 @@ would appreciate credit if this program or parts of it are used.
 #else
 #  include <fcntl.h>
 #endif
+
+#ifdef HAVE_STRREDIR_H
+#include <sys/strredir.h>
+# ifdef SRIOCSREDIR
+#  undef TIOCCONS
+# endif
+#endif
+
 #include <signal.h>
 /*#include <memory.h> - deprecated - ANSI C moves them into string.h */
 #include "string.h"
-#include <varargs.h>
+
+#ifdef HAVE_VARARGS_H
+#  include <varargs.h>
+#else
+#  ifdef HAVE_STDARG_H
+#    include <stdarg.h>
+#  endif
+#endif
+
 #include <errno.h>
 #include "exp_rename.h"
 #define EXP_DEFINE_FNS
@@ -78,7 +94,12 @@ int Exp_StringMatch();
 
 void exp_init_pty();
 
+#ifndef HAVE_TCLREGERROR
 extern char *tclRegexpError;	/* declared in tclInt.h */
+#else
+extern char *TclGetRegError();
+extern void TclRegError();
+#endif
 
 static unsigned int bufsiz = 2*EXP_MATCH_MAX;
 
@@ -197,8 +218,12 @@ when trapping, see below in child half of fork */
 		return -1;
 	}
 
-	pipe(sync_fds);
-	pipe(sync2_fds);
+	if (-1 == (pipe(sync_fds))) {
+		return -1;
+	}
+	if (-1 == (pipe(sync2_fds))) {
+		return -1;
+	}
 
 	if ((exp_pid = fork()) == -1) return(-1);
 	if (exp_pid) {
@@ -418,16 +443,31 @@ when trapping, see below in child half of fork */
 #endif /* _CRAY2 */
 #endif /* CRAY */
 
-#ifdef TIOCCONS
 	if (exp_console) {
+#ifdef SRIOCSREDIR
+		int fd;
+
+		if ((fd = open("/dev/console", O_RDONLY)) == -1) {
+			restore_error_fd
+			fprintf(stderr, "spawn %s: cannot open console, check permissions of /dev/console\n",argv[0]);
+			exit(-1);
+		}
+		if (ioctl(fd, SRIOCSREDIR, 0) == -1) {
+			restore_error_fd
+			fprintf(stderr, "spawn %s: cannot redirect console, check permissions of /dev/console\n",argv[0]);
+		}
+		close(fd);
+#endif
+
+#ifdef TIOCCONS
 		int on = 1;
 		if (ioctl(0,TIOCCONS,(char *)&on) == -1) {
 			restore_error_fd
 			fprintf(stderr, "spawn %s: cannot open console, check permissions of /dev/console\n",argv[0]);
 			exit(-1);
 		}
-	}
 #endif /* TIOCCONS */
+	}
 
 	/* tell parent that we are done setting up pty */
 	/* The actual char sent back is irrelevant. */
@@ -479,23 +519,26 @@ when trapping, see below in child half of fork */
 /* returns fd of master side of pty */
 /*VARARGS*/
 int
-exp_spawnl(va_alist)
-va_dcl
+exp_spawnl EXP_VARARGS_DEF(char *,arg1)
+/*exp_spawnl(va_alist)*/
+/*va_dcl*/
 {
 	va_list args;
 	int i;
 	char *arg, **argv;
 
-	va_start(args);
-	for (i=0;;i++) {
+	arg = EXP_VARARGS_START(char *,arg1,args);
+	/*va_start(args);*/
+	for (i=1;;i++) {
 		arg = va_arg(args,char *);
 		if (!arg) break;
 	}
 	va_end(args);
 	if (i == 0) sysreturn(EINVAL);
 	if (!(argv = (char **)malloc((i+1)*sizeof(char *)))) sysreturn(ENOMEM);
-	va_start(args);
-	for (i=0;;i++) {
+	argv[0] = EXP_VARARGS_START(char *,arg1,args);
+	/*va_start(args);*/
+	for (i=1;;i++) {
 		argv[i] = va_arg(args,char *);
 		if (!argv[i]) break;
 	}
@@ -504,6 +547,18 @@ va_dcl
 	return(i);
 }
 
+/* allow user-provided fd to be passed to expect funcs */
+int
+exp_spawnfd(fd)
+int fd;
+{
+	if (!fd_new(fd)) {
+		errno = ENOMEM;
+		return -1;
+	}
+	return fd;	
+}
+
 /* remove nulls from s.  Initially, the number of chars in s is c, */
 /* not strlen(s).  This count does not include the trailing null. */
 /* returns number of nulls removed. */
@@ -656,6 +711,7 @@ struct exp_case *ecases;
 	int old_length;		/* old buf_length */
 	int first_time = TRUE;	/* force old buffer to be tested before */
 				/* additional reads */
+	int polled = 0;		/* true if poll has caused read() to occur */
 
 	struct exp_case *ec;	/* points to current ecase */
 
@@ -692,9 +748,17 @@ struct exp_case *ecases;
 	/* compile if necessary */
 	for (ec=ecases;ec->type != exp_end;ec++) {
 		if ((ec->type == exp_regexp) && !ec->re) {
+#ifndef HAVE_TCLREGERROR
 			tclRegexpError = 0;
+#else
+			TclRegError((char *)0);
+#endif
 			if (!(ec->re = TclRegComp(ec->pattern))) {
+#ifndef HAVE_TCLREGERROR
 				fprintf(stderr,"regular expression %s is bad: %s",ec->pattern,tclRegexpError);
+#else
+				fprintf(stderr,"regular expression %s is bad: %s",ec->pattern,TclGetRegError());
+#endif
 				return_errno(EINVAL);
 			  }
 		  }
@@ -732,9 +796,6 @@ struct exp_case *ecases;
 	remtime = exp_timeout;
 
 	time(&current_time);
-	/* if user sets timeout to 0 (i.e. poll), this indicates */
-	/* we should do one read and return */
-	/* Hopefully, the buffer is big enough and will get it all! */
 	end_time = current_time + remtime;
 
 	for (;;) {
@@ -758,24 +819,39 @@ struct exp_case *ecases;
 			exp_buffer_end = exp_buffer + second_half;
 		}
 
-		if ((exp_timeout != -1) && (remtime < 0)) {
+		/*
+		 * always check first if pattern is already in buffer
+		 */
+		if (first_time) {
+			first_time = FALSE;
+			goto after_read;
+		}
+
+		/*
+		 * check for timeout
+		 */
+		if (remtime < 0 || polled) {
 			debuglog("expect: timeout\r\n");
 			exp_match_end = exp_buffer;
 			return_normally(EXP_TIMEOUT);
-		} else {
-			if (first_time) {
-				first_time = FALSE;
-				goto past_read;
-			}
-			cc = i_read(fd,fp,
-					exp_buffer_end,
-					bufsiz - buf_length,
-					remtime);
 		}
 
+		/*
+		 * if timeout == 0, indicate a poll has
+		 * occurred so that next time through loop causes timeout
+		 */
+		if (exp_timeout == 0) {
+			polled = 1;
+		}
+
+		cc = i_read(fd,fp,
+				exp_buffer_end,
+				bufsiz - buf_length,
+				remtime);
+
 		if (cc == 0) {
 			debuglog("expect: eof\r\n");
-			return_normally(EXP_EOF);		/* normal EOF */
+			return_normally(EXP_EOF);	/* normal EOF */
 		} else if (cc == -1) {			/* abnormal EOF */
 			/* ptys produce EIO upon EOF - sigh */
 			if (i_read_errno == EIO) {
@@ -820,7 +896,7 @@ struct exp_case *ecases;
 		*exp_buffer_end = '\0';
                 exp_match_end = exp_buffer;
 
-	past_read:
+	after_read:
 		debuglog("expect: does {%s} match ",exp_printify(exp_buffer));
 		/* pattern supplied */
 		for (ec=ecases;ec->type != exp_end;ec++) {
@@ -852,13 +928,22 @@ struct exp_case *ecases;
 					}
 				}
 			} else {
+#ifndef HAVE_TCLREGERROR
 				tclRegexpError = 0;
+#else
+				TclRegError((char *)0);
+#endif
 				if (TclRegExec(ec->re,exp_buffer,exp_buffer)) {
 					matched = 1;
 					exp_match = ec->re->startp[0];
 					exp_match_end = ec->re->endp[0];
+#ifndef HAVE_TCLREGERROR
 				} else if (tclRegexpError) {
 			    		fprintf(stderr,"r.e. match (pattern %s) failed: %s",ec->pattern,tclRegexpError);
+#else
+				} else if (TclGetRegError()) {
+			    		fprintf(stderr,"r.e. match (pattern %s) failed: %s",ec->pattern,TclGetRegError());
+#endif
 				}
 			}
 
@@ -869,20 +954,13 @@ struct exp_case *ecases;
 			} else debuglog("no\n");
 		}
 
-		/* indicate poll failed */
-		if (exp_timeout == 0) {
-			debuglog("timeout\r\n");
-			exp_match_end = exp_buffer;
-			return_normally(EXP_TIMEOUT);
-		}
-
-		if (exp_timeout != -1) {
+		/*
+		 * Update current time and remaining time.
+		 * Don't bother if we are waiting forever or polling.
+		 */
+		if (exp_timeout > 0) {
 			time(&current_time);
 			remtime = end_time - current_time;
-
-			/* If time has expired on the nose, remtime will be */
-			/* zero.  Since this means poll, forcibly avoid it. */
-			if (remtime == 0) remtime = -1;
 		}
 	}
  cleanup:
@@ -914,8 +992,9 @@ struct exp_case *ecases;
 
 /*VARARGS*/
 int
-exp_expectl(va_alist)
-va_dcl
+exp_expectl EXP_VARARGS_DEF(int,arg1)
+/*exp_expectl(va_alist)*/
+/*va_dcl*/
 {
 	va_list args;
 	int fd;
@@ -923,8 +1002,9 @@ va_dcl
 	int i;
 	enum exp_type type;
 
-	va_start(args);
-	fd = va_arg(args,int);
+	fd = EXP_VARARGS_START(int,arg1,args);
+	/* va_start(args);*/
+	/* fd = va_arg(args,int);*/
 	/* first just count the arg sets */
 	for (i=0;;i++) {
 		type = va_arg(args,enum exp_type);
@@ -949,8 +1029,9 @@ va_dcl
 		sysreturn(ENOMEM);
 
 	/* now set up the actual cases */
-	va_start(args);
-	va_arg(args,int);		/*COMPUTED BUT NOT USED*/
+	fd = EXP_VARARGS_START(int,arg1,args);
+	/*va_start(args);*/
+	/*va_arg(args,int);*/		/*COMPUTED BUT NOT USED*/
 	for (ec=ecases;;ec++) {
 		ec->type = va_arg(args,enum exp_type);
 		if (ec->type == exp_end) break;
@@ -976,8 +1057,9 @@ va_dcl
 }
 
 int
-exp_fexpectl(va_alist)
-va_dcl
+exp_fexpectl EXP_VARARGS_DEF(FILE *,arg1)
+/*exp_fexpectl(va_alist)*/
+/*va_dcl*/
 {
 	va_list args;
 	FILE *fp;
@@ -985,8 +1067,9 @@ va_dcl
 	int i;
 	enum exp_type type;
 
-	va_start(args);
-	fp = va_arg(args,FILE *);
+	fp = EXP_VARARGS_START(FILE *,arg1,args);
+	/*va_start(args);*/
+	/*fp = va_arg(args,FILE *);*/
 	/* first just count the arg-pairs */
 	for (i=0;;i++) {
 		type = va_arg(args,enum exp_type);
