@@ -36,6 +36,15 @@ $Date: 1993/03/03 00:21:20 $
 #include "tclRegexp.h"
 #include "exp_regexp.h"
 
+#ifndef HAVE_TCLREGERROR
+extern char *tclRegexpError;	/* declared in tclInt.h */
+#else
+extern char *TclGetRegError();
+extern void TclRegError();
+#endif
+
+#define INTER_OUT "interact_out"
+
 /*
  * tests if we are running this using a real tty
  *
@@ -104,8 +113,8 @@ static struct action *new_action();
 static int inter_eval();
 
 /* in_keymap() accepts user keystrokes and returns one of MATCH,
-CAN_MATCH, or CANT_MATCH.  These describe whether the keystrokes match
-a key sequence, and could or can't if more characters arrive.  The
+CANMATCH, or CANTMATCH.  These describe whether the keystrokes match a
+key sequence, and could or can't if more characters arrive.  The
 function assigns a matching keymap if there is a match or can-match.
 A matching keymap is assigned on can-match so we know whether to echo
 or not.
@@ -120,7 +129,7 @@ The basic idea of how this works is it does a smart sequential search.
 At each position of the input string, we attempt to match each of the
 keymaps.  If at least one matches, the first match is returned.
 
-If there is a can-match and there are more keymaps to try, we continue
+If there is a CANMATCH and there are more keymaps to try, we continue
 trying.  If there are no more keymaps to try, we stop trying and
 return with an indication of the first keymap that can match.
 
@@ -132,13 +141,14 @@ we're ready).  The other is to return can-match.
 */
 
 static int
-in_keymap(string,stringlen,keymap,km_match,match_length,skip)
+in_keymap(string,stringlen,keymap,km_match,match_length,skip,rm_nulls)
 char *string;
 int stringlen;
 struct keymap *keymap;		/* linked list of keymaps */
 struct keymap **km_match;	/* keymap that matches or can match */
 int *match_length;		/* # of chars that matched */
 int *skip;			/* # of chars to skip */
+int rm_nulls;			/* skip nulls if true */
 {
 	struct keymap *km;
 	char *ks;		/* string from a keymap */
@@ -163,7 +173,7 @@ int *skip;			/* # of chars to skip */
 /* skip over nulls - Pascal Meheut, pascal@cnam.cnam.fr 18-May-1993 */
 /*    for (start_search = string;*start_search;start_search++) {*/
     for (start_search = string;start_search<string_end;start_search++) {
-	if (*km_match) break; /* if we've already found a CAN_MATCH */
+	if (*km_match) break; /* if we've already found a CANMATCH */
 			/* don't bother starting search from positions */
 			/* further along the string */
 
@@ -191,14 +201,20 @@ int *skip;			/* # of chars to skip */
 			/* if we ran out of user-supplied characters, and */
 			/* still haven't matched, it might match if the user */
 			/* supplies more characters next time */
+
 			if (s == string_end) {
 				/* skip to next key entry, but remember */
 				/* possibility that this entry might match */
 				if (!*km_match) *km_match = km;
 				break;
 			}
-			/* skip to next key entry, if characters don't match */
-			if ((*s & 0x7f) != *ks) break;
+
+			if ((*s & 0x7f) == *ks) continue;
+			if ((*s == '\0') && rm_nulls) {
+				ks--;
+				continue;
+			}
+			break;
 		}
 	    } else {
 		/* regexp */
@@ -229,8 +245,30 @@ int *skip;			/* # of chars to skip */
 
 	if (*km_match) {
 		/* report a can-match */
+
+		char *p;
+
 		*skip = (start_search-string)-1;
+#if 0
 		*match_length = stringlen - *skip;
+#else
+		/*
+		 * there may be nulls in the string in which case
+		 * the pattern matchers can report CANMATCH when
+		 * the null is hit.  So find the null and compute
+		 * the length of the possible match.
+		 *
+		 * Later, after we squeeze out the nulls, we will
+		 * retry the match, but for now, go along with
+		 * calling it a CANMATCH
+		 */
+		p = start_search;
+		while (*p) {
+			p++;
+		}
+		*match_length = (p - start_search) + 1;
+		/*printf(" match_length = %d\n",*match_length);*/
+#endif
 		return(EXP_CANMATCH);
 	}
 
@@ -468,9 +506,11 @@ char **argv;
 	struct input **fd_to_input;	/* map from fd's to "struct input"s */
 	int *fd_list;
 	struct keymap *km;	/* ptr for above while parsing */
- 	extern char *tclRegexpError;	/* declared in tclInt.h */
+/* 	extern char *tclRegexpError;	/* declared in tclInt.h */
 	int master = EXP_SPAWN_ID_BAD;
 	char *master_string = 0;/* string representation of master */
+	int need_to_close_master = FALSE;	/* if an eof is received */
+				/* we use this to defer close until later */
 
 	int next_tty_reset = FALSE;	/* if we've seen a single -reset */
 	int next_iread = FALSE;/* if we've seen a single -iread */
@@ -518,12 +558,12 @@ char **argv;
 	int configure_count;	/* monitor reconfigure events */
 
 	if ((argc == 2) && exp_one_arg_braced(argv[1])) {
-		return(exp_eval_with_one_arg(clientData,interp,argc,argv));
+		return(exp_eval_with_one_arg(clientData,interp,argv));
 	} else if ((argc == 3) && streq(argv[1],"-brace")) {
 		char *new_argv[2];
 		new_argv[0] = argv[0];
 		new_argv[1] = argv[2];
-		return(exp_eval_with_one_arg(clientData,interp,argc,new_argv));
+		return(exp_eval_with_one_arg(clientData,interp,new_argv));
 	}
 
 	argv++;
@@ -814,6 +854,13 @@ char **argv;
 				debuglog("-timestamp is deprecated, use exp_timestamp command\r\n");
 				next_timestamp = TRUE;
 				continue;
+			} else if (exp_flageq("nobrace",arg,7)) {
+				/* nobrace does nothing but take up space */
+				/* on the command line which prevents */
+				/* us from re-expanding any command lines */
+				/* of one argument that looks like it should */
+				/* be expanded to multiple arguments. */
+				continue;
 			}
 		}
 
@@ -848,9 +895,18 @@ char **argv;
 		km->null = FALSE;
 		km->re = 0;
 		if (next_re) {
+#ifndef HAVE_TCLREGERROR
+			tclRegexpError = 0;
+#else
+			TclRegError((char *)0);
+#endif
 			if (0 == (km->re = TclRegComp(*argv))) {
 				exp_error(interp,"bad regular expression: %s",
-					tclRegexpError);
+#ifndef HAVE_TCLREGERROR
+								tclRegexpError);
+#else
+								TclGetRegError());
+#endif
 				return(TCL_ERROR);
 			}
 			next_re = FALSE;
@@ -1008,6 +1064,7 @@ char **argv;
 			/* refined by matcher. */
 		int cc;	/* chars count from read() */
 		int m;	/* master */
+		int m_out; /* where master echoes to */
 		struct action *action = 0;
 		time_t previous_time;
 		time_t current_time;
@@ -1097,6 +1154,17 @@ char **argv;
 
 		switch (rc) {
 		case EXP_DATA_NEW:
+			if (u->size == u->msize) {
+			    /* In theory, interact could be invoked when this situation */
+			    /* already exists, hence the "probably" in the warning below */
+
+			    debuglog("WARNING: interact buffer is full, probably because your\r\n");
+			    debuglog("patterns have matched all of it but require more chars\r\n");
+			    debuglog("in order to complete the match.\r\n");
+			    debuglog("Dumping first half of buffer in order to continue\r\n");
+			    debuglog("Recommend you enlarge the buffer or fix your patterns.\r\n");
+			    exp_buffer_shuffle(interp,u,0,INTER_OUT,"interact");
+		        }
 			cc = read(m,	u->buffer + u->size,
 					u->msize - u->size);
 			if (cc > 0) {
@@ -1121,20 +1189,33 @@ char **argv;
 				}
 				break;
 			}
-			/*FALLTHRU*/
 
+			rc = EXP_EOF;
 			/* Most systems have read() return 0, allowing */
 			/* control to fall thru and into this code.  On some */
 			/* systems (currently HP and new SGI), read() does */
 			/* see eof, and it must be detected earlier.  Then */
 			/* control jumps directly to this EXP_EOF label. */
+
+			/*FALLTHRU*/
 		case EXP_EOF:
 			action = inp->action_eof;
 			attempt_match = FALSE;
 			skip = u->size;
-			rc = EXP_EOF;
 			debuglog("interact: received eof from spawn_id %d\r\n",m);
+			/* actual close is done later so that we have a */
+			/* chance to flush out any remaining characters */
+			need_to_close_master = TRUE;
+
+#if EOF_SO
+			/* should really check for remaining chars and */
+			/* flush them but this will only happen in the */
+			/* unlikely scenario that there are partially */
+			/* matched buffered chars. */
+			/* So for now, indicate no chars to skip. */
+			skip = 0;
 			exp_close(interp,m);
+#endif
 			break;
 		case EXP_DATA_OLD:
 			cc = 0;
@@ -1150,14 +1231,13 @@ char **argv;
 
 		if (attempt_match) {
 			rc = in_keymap(u->buffer,u->size,inp->keymap,
-				&km,&match_length,&skip);
+				&km,&match_length,&skip,u->rm_nulls);
 		} else {
 			attempt_match = TRUE;
 		}
 
 		/* put regexp result in variables */
 		if (km && km->re) {
-#define INTER_OUT "interact_out"
 #define out(var,val)  debuglog("expect: set %s(%s) \"%s\"\r\n",INTER_OUT,var, \
 						dprintify(val)); \
 		    Tcl_SetVar2(interp,INTER_OUT,var,val,0);
@@ -1196,29 +1276,34 @@ char **argv;
 			}
 		}
 
-		/* dispose of chars that should be skipped */
+		/*
+		 * dispose of chars that should be skipped
+		 * i.e., chars that cannot possibly be part of a match.
+		 */
 		
-		/* skip is chars not involved in match */
-		/* print is with chars involved in match */
+		/* "skip" is count of chars not involved in match */
+		/* "print" is count with chars involved in match */
 
 		if (km && km->writethru) {
 			print = skip + match_length;
 		} else print = skip;
 
-		/* figure out if we should echo any chars */
+		/*
+		 * echo chars if appropriate
+		 */
 		if (km && km->echo) {
 			int seen;	/* either printed or echoed */
 
 			/* echo to stdout rather than stdin */
-			if (m == 0) m = 1;
+			m_out = (m == 0)?1:m;
 
 			/* write is unlikely to fail, since we just read */
 			/* from same descriptor */
 			seen = u->printed + u->echoed;
 			if (skip >= seen) {
-				write(m,u->buffer+skip,match_length);
+				write(m_out,u->buffer+skip,match_length);
 			} else if ((match_length + skip - seen) > 0) {
-				write(m,u->buffer+seen,match_length+skip-seen);
+				write(m_out,u->buffer+seen,match_length+skip-seen);
 			}
 			u->echoed = match_length + skip - u->printed;
 		}
@@ -1314,11 +1399,17 @@ char **argv;
 			}
 		}
 
+#if EOF_SO
 		/* as long as buffer is still around, null terminate it */
 		if (rc != EXP_EOF) {
 			u->buffer[u->size] = '\0';
 			u->lower [u->size] = '\0';
 		}
+#else
+		u->buffer[u->size] = '\0';
+		u->lower [u->size] = '\0';
+#endif
+
 		/* now update printed based on total amount skipped */
 
 		u->printed -= skip;
@@ -1335,6 +1426,11 @@ char **argv;
 			if (skip >= oldprinted + u->echoed) u->echoed = 0;
 		}
 
+		if (rc == EXP_EOF) {
+			exp_close(interp,m);
+			need_to_close_master = FALSE;
+		}
+
 		if (action) {
 got_action:
 			change = (action && action->tty_reset);
@@ -2041,6 +2137,8 @@ got_action:
 	}
 #endif /* SIMPLE_EVENT */
 
+	if (need_to_close_master) exp_close(interp,master);
+
 	if (tty_changed) exp_tty_set(interp,&tty_old,was_raw,was_echo);
 	if (oldargv) ckfree((char *)argv);
 	if (fd_list) ckfree((char *)fd_list);
