@@ -1,13 +1,18 @@
 /* exp_tty.c - tty support routines */
 
+#include "expect_cf.h"
 #include <stdio.h>
 #include <signal.h>
+#include <sys/types.h>
+
+#ifdef HAVE_SYS_WAIT_H
+#include <sys/wait.h>
+#endif
 
 #if defined(SIGCLD) && !defined(SIGCHLD)
 #define SIGCHLD SIGCLD
 #endif
 
-#include "expect_cf.h"
 #include "tcl.h"
 #include "exp_prog.h"
 #include "exp_rename.h"
@@ -22,7 +27,9 @@ int exp_ioctled_devtty = FALSE;
 int exp_stdin_is_tty;
 int exp_stdout_is_tty;
 
-static exp_tty tty_current, tty_cooked;
+/*static*/ extern exp_tty exp_tty_current, exp_tty_cooked;
+#define tty_current exp_tty_current
+#define tty_cooked exp_tty_cooked
 
 int
 exp_israw()
@@ -53,7 +60,9 @@ int set;
 		tty_current.c_iflag = tty_cooked.c_iflag;
 		tty_current.c_oflag = tty_cooked.c_oflag;
 /*		tty_current.c_lflag = tty_cooked.c_lflag;*/
-		tty_current.c_lflag = tty_cooked.c_lflag & ~ECHO;
+/* attempt 2	tty_current.c_lflag = tty_cooked.c_lflag & ~ECHO;*/
+		/* retain current echo setting */
+		tty_current.c_lflag = (tty_cooked.c_lflag & ~ECHO) | (tty_current.c_lflag & ECHO);
 		tty_current.c_cc[VMIN] = tty_cooked.c_cc[VMIN];
 		tty_current.c_cc[VTIME] = tty_cooked.c_cc[VTIME];
 #else
@@ -90,7 +99,7 @@ int
 exp_tty_set_simple(tty)
 exp_tty *tty;
 {
-#ifdef POSIX
+#ifdef HAVE_TCSETATTR
 	return(tcsetattr(exp_dev_tty, TCSADRAIN,tty));
 #else
 	return(ioctl    (exp_dev_tty, TCSETSW  ,tty));
@@ -101,7 +110,7 @@ int
 exp_tty_get_simple(tty)
 exp_tty *tty;
 {
-#ifdef POSIX
+#ifdef HAVE_TCSETATTR
 	return(tcgetattr(exp_dev_tty,         tty));
 #else
 	return(ioctl    (exp_dev_tty, TCGETS, tty));
@@ -203,19 +212,10 @@ exp_update_current_from_real_tty() {
 #endif
 
 void
-exp_init_tty()
+exp_init_stdio()
 {
-	extern exp_tty exp_tty_original;
-
-	/* save original user tty-setting in 'cooked', just in case user */
-	/* asks for it without earlier telling us what cooked means to them */
-	tty_cooked = exp_tty_original;
-
-	/* save our current idea of the terminal settings */
-	tty_current = exp_tty_original;
-
 	exp_stdin_is_tty = isatty(0);
-	exp_stdout_is_tty = isatty(0);
+	exp_stdout_is_tty = isatty(1);
 
 	setbuf(stdout,(char *)0);	/* unbuffer stdout */
 }
@@ -231,7 +231,7 @@ int fd;
 #else
 # ifdef TIOCSBRK
 	ioctl(fd,TIOCSBRK,0);
-	exp_usleep(interp,250000); /* sleep for at least a quarter of a second */
+	exp_dsleep(interp,0.25); /* sleep for at least a quarter of a second */
 	ioctl(fd,TIOCCBRK,0);
 # else
 	/* dunno how to do this - ignore */
@@ -535,14 +535,17 @@ char **argv;
 	RETSIGTYPE (*old)();	/* save old sigalarm handler */
 #define MAX_ARGLIST 10240
 	int i;
-	int waitStatus;
+
+	WAIT_STATUS_TYPE waitStatus;
+	int systemStatus
+;
 	int abnormalExit = FALSE;
 	char buf[MAX_ARGLIST];
 	char *bufp = buf;
 	int total_len = 0, arg_len;
 
 	int stty_args_recognized = TRUE;
-	int stty_cmd = FALSE;
+	int cmd_is_stty = FALSE;
 	int cooked = FALSE;
 	int was_raw, was_echo;
 
@@ -551,12 +554,12 @@ char **argv;
 	if (streq(argv[1],"stty")) {
 		debuglog("system stty is deprecated, use stty\r\n");
 
-		stty_cmd = TRUE;
+		cmd_is_stty = TRUE;
 		was_raw = exp_israw();
 		was_echo = exp_isecho();
 	}
 
-	if (argc > 2 && stty_cmd) {
+	if (argc > 2 && cmd_is_stty) {
 		exp_ioctled_devtty = TRUE;
 
 		for (i=2;i<argc;i++) {
@@ -573,9 +576,10 @@ char **argv;
 				exp_tty_echo(-1);
 			} else stty_args_recognized = FALSE;
 		}
+
 		/* if unknown args, fall thru and let real stty have a go */
 		if (stty_args_recognized) {
-#ifdef POSIX
+#ifdef HAVE_TCSETATTR
  			if (tcsetattr(exp_dev_tty,TCSADRAIN, &tty_current) == -1) {
 #else
 		        if (ioctl(exp_dev_tty, TCSETSW, &tty_current) == -1) {
@@ -588,7 +592,7 @@ char **argv;
 			    exp_error(interp,"system stty: ioctl(user): %s\r\n",Tcl_PosixError(interp));
 			    return(TCL_ERROR);
 			}
-			if (stty_cmd) {
+			if (cmd_is_stty) {
 				sprintf(interp->result,"%sraw %secho",
 					(was_raw?"":"-"),
 					(was_echo?"":"-"));
@@ -613,13 +617,19 @@ char **argv;
 
 	*(bufp-1) = '\0';
 	old = signal(SIGCHLD, SIG_DFL);
-	waitStatus = system(buf);
+	systemStatus = system(buf);
 	signal(SIGCHLD, old);	/* restore signal handler */
 	debuglog("system(%s) = %d\r\n",buf,i);
 
+	if (systemStatus == -1) {
+		exp_error(interp,Tcl_PosixError(interp));
+		return TCL_ERROR;
+	}
+	*(int *)&waitStatus = systemStatus;
+
 	if (!stty_args_recognized) {
 		/* find out what weird options user asked for */
-#ifdef POSIX
+#ifdef HAVE_TCSETATTR
 		if (tcgetattr(exp_dev_tty, &tty_current) == -1) {
 #else
 	        if (ioctl(exp_dev_tty, TCGETS, &tty_current) == -1) {
@@ -633,7 +643,7 @@ char **argv;
 		}
 	}
 
-	if (stty_cmd) {
+	if (cmd_is_stty) {
 		sprintf(interp->result,"%sraw %secho",
 			(was_raw?"":"-"),
 			(was_echo?"":"-"));
@@ -643,30 +653,40 @@ char **argv;
 /* we can't include the whole thing because it depends on other macros */
 /* that come out of Tcl's Makefile, sigh */
 
+#if 0
+
+#undef WIFEXITED
 #ifndef WIFEXITED
 #   define WIFEXITED(stat)  (((*((int *) &(stat))) & 0xff) == 0)
 #endif
 
+#undef WEXITSTATUS
 #ifndef WEXITSTATUS
 #   define WEXITSTATUS(stat) (((*((int *) &(stat))) >> 8) & 0xff)
 #endif
 
+#undef WIFSIGNALED
 #ifndef WIFSIGNALED
 #   define WIFSIGNALED(stat) (((*((int *) &(stat)))) && ((*((int *) &(stat))) == ((*((int *) &(stat))) & 0x00ff)))
 #endif
 
+#undef WTERMSIG
 #ifndef WTERMSIG
 #   define WTERMSIG(stat)    ((*((int *) &(stat))) & 0x7f)
 #endif
 
+#undef WIFSTOPPED
 #ifndef WIFSTOPPED
 #   define WIFSTOPPED(stat)  (((*((int *) &(stat))) & 0xff) == 0177)
 #endif
 
+#undef WSTOPSIG
 #ifndef WSTOPSIG
 #   define WSTOPSIG(stat)    (((*((int *) &(stat))) >> 8) & 0xff)
 #endif
 
+#endif /* 0 */
+
 /* stolen from Tcl.    Again, this is embedded in another routine */
 /* (CleanupChildren in tclUnixAZ.c) that we can't use directly. */
 
