@@ -26,7 +26,9 @@ would appreciate credit if this program or parts of it are used.
 #endif
 
 #include "tcl.h"
+#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION < 4
 extern char *tclRegexpError;	/* declared in tclInt.h */
+#endif
 
 #include "string.h"
 
@@ -98,6 +100,8 @@ static
 struct exp_cmd_descriptor {
 	int cmdtype;			/* bg, before, after */
 	int duration;			/* permanent or temporary */
+	int timeout_specified_by_flag;	/* if -timeout flag used */
+	int timeout;			/* timeout period if flag used */
 	struct exp_cases_descriptor ecd;
 	struct exp_i *i_list;
 } exp_cmds[3];
@@ -344,36 +348,41 @@ char *p;
 /* returns TCL_whatever */
 /*ARGSUSED*/
 int
-exp_eval_with_one_arg(clientData,interp,argc,argv)
+exp_eval_with_one_arg(clientData,interp,argv)
 ClientData clientData;
 Tcl_Interp *interp;
-int argc;
 char **argv;
 {
 	char *buf;
 	int rc;
 	char *a;
 
-	/* + 2 is for blank separating cmd and null at end */
-	buf = ckalloc(strlen(argv[0]) + strlen(argv[1]) + 2);
+	/* + 11 is for " -nobrace " and null at end */
+	buf = ckalloc(strlen(argv[0]) + strlen(argv[1]) + 11);
+	/* recreate statement (with -nobrace to prevent recursion) */
+	sprintf(buf,"%s -nobrace %s",argv[0],argv[1]);
 
-	/* replace top-level newlines with blanks */
-	for (a=argv[1];*a;) {
+	/*
+	 * replace top-level newlines with blanks
+	 */
+
+	/* Should only be necessary to run over argv[1] and then sprintf */
+	/* that into the buffer, but the ICEM guys insist that writing */
+	/* back over the original arguments makes their Tcl compiler very */
+	/* unhappy. */
+	for (a=buf;*a;) {
 		extern char *TclWordEnd();
 
 		for (;isspace(*a);a++) {
 			if (*a == '\n') *a = ' ';
 		}
-#if TCL_MINOR_VERSION == 0
+#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION == 0
 		a = TclWordEnd(a,0)+1;
 #else
 		a = TclWordEnd(a,0,(int *)0)+1;
 #endif
 	}
 
-	/* recreate statement */
-	sprintf(buf,"%s %s",argv[0],argv[1]);
-
 	rc = Tcl_Eval(interp,buf);
 
 	ckfree(buf);
@@ -447,6 +456,8 @@ char **argv;
 	argv++;
 	argc--;
 
+	eg->timeout_specified_by_flag = FALSE;
+
 	ecase_clear(&ec);
 
 	/* Allocate an array to store the ecases.  Force array even if 0 */
@@ -482,10 +493,18 @@ char **argv;
 			} else if (exp_flageq("regexp",arg,2)) {
 				i++;
 				ec.use = PAT_RE;
+#ifndef HAVE_TCLREGERROR
 				tclRegexpError = 0;
+#else
+				TclRegError((char *)0);
+#endif
 				if (!(ec.re = TclRegComp(argv[i]))) {
 					exp_error(interp,"bad regular expression: %s",
+#ifndef HAVE_TCLREGERROR
 								tclRegexpError);
+#else
+								TclGetRegError());
+#endif
 					goto error;
 				}
 			} else if (exp_flageq("exact",arg,2)) {
@@ -526,6 +545,23 @@ char **argv;
 			} else if (exp_flageq("timestamp",arg,2)) {
 				ec.timestamp = TRUE;
 				continue;
+			} else if (exp_flageq("timeout",arg,2)) {
+				i++;
+				if (i>=argc) {
+					exp_error(interp,"-timeout requires following # of seconds");
+					goto error;
+				}
+
+				eg->timeout = atoi(argv[i]);
+				eg->timeout_specified_by_flag = TRUE;
+				continue;
+			} else if (exp_flageq("nobrace",arg,7)) {
+				/* nobrace does nothing but take up space */
+				/* on the command line which prevents */
+				/* us from re-expanding any command lines */
+				/* of one argument that looks like it should */
+				/* be expanded to multiple arguments. */
+				continue;
 			} else {
 				exp_error(interp,"usage: unrecognized flag <%s>",arg);
 				goto error;
@@ -646,7 +682,11 @@ char *suffix;
 
 	if (e->use == PAT_RE) {
 		debuglog("\"%s\"? ",dprintify(e->pat));
+#ifndef HAVE_TCLREGERROR
 		tclRegexpError = 0;
+#else
+		TclRegError((char *)0);
+#endif
 		if (buffer && TclRegExec(e->re,buffer,buffer)) {
 			o->e = e;
 			o->match = e->re->endp[0]-buffer;
@@ -656,8 +696,13 @@ char *suffix;
 			return(EXP_MATCH);
 		} else {
 			debuglog(no);
+#ifndef HAVE_TCLREGERROR
 			if (tclRegexpError) {
 			    exp_error(interp,"-re failed: %s",tclRegexpError);
+#else
+			if (TclGetRegError()) {
+			    exp_error(interp,"-re failed: %s",TclGetRegError());
+#endif
 			    return(EXP_TCLERROR);
 		        }
 		    }
@@ -1170,12 +1215,12 @@ char **argv;
 	struct exp_cmd_descriptor *ecmd = (struct exp_cmd_descriptor *) clientData;
 
 	if ((argc == 2) && exp_one_arg_braced(argv[1])) {
-		return(exp_eval_with_one_arg(clientData,interp,argc,argv));
+		return(exp_eval_with_one_arg(clientData,interp,argv));
 	} else if ((argc == 3) && streq(argv[1],"-brace")) {
 		char *new_argv[2];
 		new_argv[0] = argv[0];
 		new_argv[1] = argv[2];
-		return(exp_eval_with_one_arg(clientData,interp,argc,new_argv));
+		return(exp_eval_with_one_arg(clientData,interp,new_argv));
 	}
 
 	if (argc > 1 && (argv[1][0] == '-')) {
@@ -1414,7 +1459,7 @@ struct exp_f *f;
 		}
 		f->msize = new_msize;
 		f->buffer[f->size] = '\0';
-
+		f->lower[f->size] = '\0';
 	}
 }
 
@@ -1504,6 +1549,9 @@ int key;
 		if (i_read_errno == EIO) {
 			/* Sun, Cray, BSD, and others */
 			cc = EXP_EOF;
+		} else if (i_read_errno == EINVAL) {
+			/* Solaris 2.4 occasionally returns this */
+			cc = EXP_EOF;
 		} else {
 			if (i_read_errno == EBADF) {
 				exp_error(interp,"bad spawn_id (process died earlier?)");
@@ -1566,11 +1614,13 @@ static int i_read_cc;
 
 /* when buffer fills, copy second half over first and */
 /* continue, so we can do matches over multiple buffers */
-static void
-buffer_shuffle(interp,f,save_flags)
+void
+exp_buffer_shuffle(interp,f,save_flags,array_name,caller_name)
 Tcl_Interp *interp;
 struct exp_f *f;
 int save_flags;
+char *array_name;
+char *caller_name;
 {
 	char spawn_id[10];	/* enough for a %d */
 	char match_char;	/* place to hold char temporarily */
@@ -1582,16 +1632,20 @@ int save_flags;
 	/*
 	 * allow user to see data we are discarding
 	 */
-#define out(i,val,flags) \
-		    debuglog("expect: set %s(%s) \"%s\"\r\n",EXPECT_OUT,i, \
-						dprintify(val)); \
-		    Tcl_SetVar2(interp,EXPECT_OUT,i,val,flags);
+
 	sprintf(spawn_id,"%d",f-exp_fs);
-	out("spawn_id",spawn_id,save_flags);
+	debuglog("%s: set %s(spawn_id) \"%s\"\r\n",
+		 caller_name,array_name,dprintify(spawn_id));
+	Tcl_SetVar2(interp,array_name,"spawn_id",spawn_id,save_flags);
+
 	/* temporarily null-terminate buffer in middle */
 	match_char = f->buffer[first_half];
 	f->buffer[first_half] = 0;
-	out("buffer",f->buffer,save_flags);
+
+	debuglog("%s: set %s(buffer) \"%s\"\r\n",
+		 caller_name,array_name,dprintify(f->buffer));
+	Tcl_SetVar2(interp,array_name,"buffer",f->buffer,save_flags);
+
 	/* remove middle-null-terminator */
 	f->buffer[first_half] = match_char;
 
@@ -1601,7 +1655,6 @@ int save_flags;
 	f->printed -= first_half;
 	if (f->printed < 0) f->printed = 0;
 }
-#undef out
 
 /* map EXP_ style return value to TCL_ style return value */
 /* not defined to work on TCL_OK */
@@ -1652,7 +1705,8 @@ int save_flags;
 	int cc = EXP_TIMEOUT;
 
 	f = exp_fs + m;
-	if (f->size == f->msize) buffer_shuffle(interp,f,save_flags);
+	if (f->size == f->msize) 
+		exp_buffer_shuffle(interp,f,save_flags,EXPECT_OUT,"expect");
 
 #ifdef SIMPLE_EVENT
  restart:
@@ -2201,12 +2255,12 @@ char **argv;
 	int reset_timer;	/* should timer be reset after continue? */
 
 	if ((argc == 2) && exp_one_arg_braced(argv[1])) {
-		return(exp_eval_with_one_arg(clientData,interp,argc,argv));
+		return(exp_eval_with_one_arg(clientData,interp,argv));
 	} else if ((argc == 3) && streq(argv[1],"-brace")) {
 		char *new_argv[2];
 		new_argv[0] = argv[0];
 		new_argv[1] = argv[2];
-		return(exp_eval_with_one_arg(clientData,interp,argc,new_argv));
+		return(exp_eval_with_one_arg(clientData,interp,new_argv));
 	}
 
 	time(&start_time_total);
@@ -2258,8 +2312,12 @@ char **argv;
 	if (first_time) first_time = 0;
 	else time(&start_time);
 
-	/* get the latest timeout */
-	timeout = get_timeout(interp);
+	if (eg.timeout_specified_by_flag) {
+		timeout = eg.timeout;
+	} else {
+		/* get the latest timeout */
+		timeout = get_timeout(interp);
+	}
 
 	key = expect_key++;
 
